/*
 * Copyright (c) Members of the EGEE Collaboration. 2009-2010.
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <iostream>
#include <vector>
#include <string>
#include <limits>
#include <stdlib.h>

#include "saga/saga.hpp"
#include "saga/saga/sd.hpp"

bool compare_nocase(const std::string& first, const std::string& second) {
    size_t len = 0;
    while ((len < first.length()) && (len < second.length())) {
        if (std::tolower(first[len]) < std::tolower(second[len])) {
            return true;
        }

        else if (std::tolower(first[len]) > tolower(second[len])) {
            return false;
        }
        ++len;
    }

    if (first.length() < second.length()) {
        return true;
    }

    else {
        return false;
    }
}

/**
 * Function to return value from a string. A check is made that there is nothing else after valid input apart from white space.
 */
template<class T>
bool from_string(T& t, const std::string& s) {
    std::istringstream iss(s);
    if ((iss >> t).fail()) {
        return false;
    }
    int eof;
    iss >> eof;
    return iss.fail() && iss.eof();
}

/*
 * Get next parameter safely
 */
std::string getNext(int & optind, int argc, char *argv[]) {
    if (optind >= argc-1) {
        std::cerr << argv[optind] << " must be followed by a value" << std::endl;
        exit(1);
    }
    ++optind;
    return argv[optind];
}

/**
 * Main program
 */
int main(int argc, char *argv[]) {
    setenv("SAGA_LOCATION","/usr",1);
    saga::context ctx("glite");
    saga::session s;
    s.add_context(ctx);

    std::string svcFilter = "";
    std::string authzFilter = "";
    std::string dataFilter = "";
    std::string serverUrl = "";
    bool haveAuthzFilter = false;
    bool sa = false;
    bool da = false;
    int maxCount = std::numeric_limits<int>::max();

    int optind = 1;
    while ((optind < argc) && (argv[optind][0] == '-')) {
        std::string sw = argv[optind];

        if (sw == "--sf") {
            svcFilter = getNext(optind, argc, argv);
        } else if (sw == "--af") {
            authzFilter = getNext(optind, argc, argv);
            haveAuthzFilter = true;
        } else if (sw == "--df") {
            dataFilter = getNext(optind, argc, argv);
        } else if (sw == "--url") {
            serverUrl = getNext(optind, argc, argv);
        } else if (sw == "--da") {
            da = true;
        } else if (sw == "--sa") {
            sa = true;
        } else if (sw == "-n" || sw == "--count") {
            std::string val = getNext(optind, argc, argv);
            if (!from_string<int> (maxCount, val) || maxCount <= 0) {
                std::cerr << "Count must be a positive integer and not: " << argv[optind] << std::endl;
                exit(1);
            }
        } else if (sw == "-h" || sw == "--help") {
            std::cout << std::endl;
            std::cout << "Usage: " << argv[0] << " [options]" << std::endl;
            std::cout << "Options:" << std::endl;
            std::cout << "   --sf <service filter> the default is no filtering on service attributes" << std::endl;
            std::cout << "   --df <data filter>    the default is no filtering on service data" << std::endl;
            std::cout << "   --af <authz filter>   the default is derived from the glite context" << std::endl;
            std::cout << "   --count, -n           the number of services to display - default unlimited" << std::endl;
            std::cout << "   --sa                  display service attributes " << std::endl;
            std::cout << "   --da                  display data attributes" << std::endl;
            std::cout << "   --url <server url>    overrides any environment variables" << std::endl;
            std::cout << "   --help, -h            print this message" << std::endl;
            std::cout << std::endl;
            return 0;
        } else {
            std::cerr << "Unknown switch: " << argv[optind] << std::endl;
            std::cerr << "Run with -h for help" << std::endl;
            exit(1);
        }
        optind++;
    }

    saga::sd::discoverer sd(s, serverUrl);
    std::vector<saga::sd::service_description> services;

    try {
        if (haveAuthzFilter) {
            services = sd.list_services(svcFilter, dataFilter, authzFilter);
        } else {
            services = sd.list_services(svcFilter, dataFilter);
        }
    }

    catch (saga::exception& e) {
        std::cerr << e.get_message() << std::endl << "Exiting." << std::endl;
        exit(1);
    }

    catch (std::exception& e) {
        std::cerr << "Caught std::exception: " << e.what() << std::endl;
        exit(1);
    }

    catch (...) {
        std::cerr << "Caught something unexpected." << std::endl;
        exit(1);
    }

    std::vector<saga::sd::service_description>::iterator servicesIter;
    std::vector<saga::sd::service_description>::const_iterator servicesEnd = services.end();

    for (servicesIter = services.begin(); servicesIter != servicesEnd && maxCount > 0; ++servicesIter, --maxCount) {
        using namespace saga::sd::attributes;

        if (da || sa) {
            std::cout << std::endl;
        }

        std::vector<std::string> attribNames = servicesIter->list_attributes();

        // Deal with the URL if it exists - it should!
        std::vector<std::string>::iterator pos = find(attribNames.begin(), attribNames.end(), "Url");
        if (pos == attribNames.end()) {
            std::cout << "Url : is not known" << std::endl;
        } else {
            std::cout << "Url : " << servicesIter->get_attribute("Url") << std::endl;
            attribNames.erase(pos);
        }

        if (sa) {
            //Sort our array of attribute names so they come out in a decent order
            sort(attribNames.begin(), attribNames.end(), compare_nocase);

            for (std::vector<std::string>::const_iterator attribNamesIter = attribNames.begin(); attribNamesIter
                    != attribNames.end(); ++attribNamesIter) {
                if (!servicesIter->attribute_is_vector(*attribNamesIter)) {
                    std::string attribValue = servicesIter->get_attribute(*attribNamesIter);
                    std::cout << "Service " << *attribNamesIter << " : " << attribValue << std::endl;
                }

                else {
                    std::vector<std::string> attribValues = servicesIter->get_vector_attribute(*attribNamesIter);
                    std::vector<std::string>::const_iterator attribValuesIter;
                    std::vector<std::string>::const_iterator attribValuesEnd = attribValues.end();
                    for (attribValuesIter = attribValues.begin(); attribValuesIter != attribValuesEnd; ++attribValuesIter) {
                        std::cout << "Service " << *attribNamesIter << " : " << *attribValuesIter << std::endl;
                    }
                }
            }
        }

        if (da) {
            saga::sd::service_data data;
            try {
                data = servicesIter->get_data();
            }

            //Not a lot we can do here so just skip it
            catch (saga::exception& e) {
                std::cerr << "EXCEPTION: " << e.get_message() << std::endl;
            }

            catch (std::exception& e) {
                std::cerr << "Caught std::exception: " << e.what() << std::endl;
            }

            catch (...) {
                std::cerr << "Caught something unexpected." << std::endl;
            }

            std::vector<std::string> dataNames = data.list_attributes();

            //Sort our array of data attribute names so they come out in a decent order
            sort(dataNames.begin(), dataNames.end(), compare_nocase);

            std::vector<std::string>::const_iterator dataIter;
            std::vector<std::string>::const_iterator dataEnd = dataNames.end();

            for (dataIter = dataNames.begin(); dataIter != dataEnd; ++dataIter) {
                if (!data.attribute_is_vector(*dataIter)) {
                    std::string dataValue;
                    dataValue = data.get_attribute(*dataIter);
                    std::cout << "Data " << *dataIter << " : " << dataValue << std::endl;
                }

                else {
                    std::vector<std::string> dataValues = data.get_vector_attribute(*dataIter);
                    std::vector<std::string>::const_iterator dataValuesIter;
                    std::vector<std::string>::const_iterator dataValuesEnd = dataValues.end();
                    for (dataValuesIter = dataValues.begin(); dataValuesIter != dataValuesEnd; ++dataValuesIter) {
                        std::cout << "Data " << *dataIter << " : " << *dataValuesIter << std::endl;
                    }
                }
            }
        }
    }
    if (da || sa) {
        std::cout << std::endl;
    }
    std::cout << services.size() << " services matched the query" << std::endl;
    if (da || sa) {
        std::cout << std::endl;
    }

    return 0;
}

